import asyncio
from abc import ABC
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, TypedDict

from pydantic import validate_arguments

from ..cache import cache
from ..core import BaseClient
from ..encoders import encode_query_params
from ..errors import ServerError
from ..retry import retry_policy

PATH = "/users/{user_uuid}/institutions/{institution_id}/alerts"

Alert = Dict[str, Any]


class AlertsResponse(TypedDict):
    page: int
    num_pages: int
    alerts: List[Alert]


class BaseAlertsResource(ABC):
    def __init__(self, client: BaseClient):
        self._client = client

    @retry_policy(
        max_retry_count=4,
        max_retry_interval_seconds=10,
        retryable_exceptions=[ServerError],
    )
    async def _get_page(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
        page: int = 1,
    ) -> AlertsResponse:
        async with self._client.session() as session:
            response = await session.get(
                PATH.format(user_uuid=user_uuid, institution_id=institution_id),
                params=encode_query_params(
                    utc_starttime=utc_starttime,
                    utc_endtime=utc_endtime,
                    labels=labels,
                    page=page,
                ),
            )

        if response.status_code == 500:
            raise ServerError(response.text)

        assert response.status_code == 200, response.text

        alerts_response: AlertsResponse = response.json()
        return alerts_response

    @validate_arguments
    @cache
    async def _get(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
        page: Optional[int] = None,
    ) -> List[Alert]:
        response = await self._get_page(
            user_uuid=user_uuid,
            institution_id=institution_id,
            utc_starttime=utc_starttime,
            utc_endtime=utc_endtime,
            labels=labels,
            page=page or 1,
        )
        max_pages = response["num_pages"]
        if not page and max_pages > 1:
            coroutines = [
                self._get_page(
                    user_uuid=user_uuid,
                    institution_id=institution_id,
                    utc_starttime=utc_starttime,
                    utc_endtime=utc_endtime,
                    labels=labels,
                    page=page + 2,
                )
                for page in range(max_pages - 1)
            ]
            response_pages = await asyncio.gather(*coroutines)
            responses = (response, *response_pages)
        else:
            responses = (response,)

        return [alert for response in responses for alert in response["alerts"]]


class AsyncAlertsResource(BaseAlertsResource):
    async def get(
        self,
        user_uuid: str,
        institution_id: str,
        *,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
        page: Optional[int] = None,
    ) -> List[Alert]:
        return await self._get(
            user_uuid=user_uuid,
            institution_id=institution_id,
            utc_starttime=utc_starttime,
            utc_endtime=utc_endtime,
            labels=labels,
            page=page,
        )


class SyncAlertsResource(BaseAlertsResource):
    def get(
        self,
        user_uuid: str,
        institution_id: str,
        *,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
        page: Optional[int] = None,
    ) -> List[Alert]:
        return self._client.run(
            self._get(
                user_uuid=user_uuid,
                institution_id=institution_id,
                utc_starttime=utc_starttime,
                utc_endtime=utc_endtime,
                labels=labels,
                page=page,
            )
        )



"""
Utilities to implement functional retry policies.

Examples:
    @retry_policy()
    def hello_world():
        return "Hello world"
"""

import logging
import time
from functools import partial, wraps
from typing import Callable, Sequence, Type, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


logger = logging.getLogger(__name__)

class RetryLimitExceeded(Exception):
    """Function raised more exceptions than allowed."""


def execute_with_retry_policy(
    function: Callable[[], T],
    min_retry_interval_seconds: float = 1,
    max_retry_interval_seconds: float = 60,
    max_retry_count: int = 10,
    retryable_exceptions: Sequence[Type[Exception]] = (Exception,),
) -> T:
    """Invoke given function and manage retries for configured exceptions.

    If the function requires positional or keyword arguments, use
    functools.partial to construct a callable with predefined arguments.
    """
    retriable_exceptions_set = set(retryable_exceptions)
    retry_count = 0
    retry_interval_seconds = min_retry_interval_seconds

    while True:
        try:
            return function()
        except Exception as e:  # pylint: disable=broad-except
            if type(e) not in retriable_exceptions_set:
                raise

            if retry_count >= max_retry_count:
                # Construct dynamically created exception which subclasses from
                # RetryLimitExceeded and the exception raised from the callable
                # to preserve exception handling behavior of the caller.
                raise type("RetryFailed", (RetryLimitExceeded, type(e)), {})

            logger.warning("Retrying exception: %s", repr(e))
            time.sleep(retry_interval_seconds)

            retry_count += 1
            retry_interval_seconds = min(
                retry_interval_seconds * 2, max_retry_interval_seconds
            )


def retry_policy(
    min_retry_interval_seconds: float = 1,
    max_retry_interval_seconds: float = 60,
    max_retry_count: int = 10,
    retryable_exceptions: Sequence[Type[Exception]] = (Exception,),
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorate function to execute in managed retry environment."""

    def function_with_retry_policy(function: Callable[P, T]) -> Callable[P, T]:
        @wraps(function)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            result = execute_with_retry_policy(
                partial(function, *args, **kwargs),
                min_retry_interval_seconds=min_retry_interval_seconds,
                max_retry_interval_seconds=max_retry_interval_seconds,
                max_retry_count=max_retry_count,
                retryable_exceptions=retryable_exceptions,
            )
            return result

        return wrapper

    return function_with_retry_policy
